/* Nastaven kurzoru */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>
#include <X11/cursorfont.h> /* Konstanty pro XCreateFontCursor() */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "Xlib_icon.h"
#include "cursor_source.h"
#include "cursor_mask.h"

/*
 * Globln pouvan promnn
 */
char *progname; /* Jmno programu z argv[0] */

char *display_name; /* Jmno pouvanho displeje */
Display *display; /* Struktura popisujc displej */
int screen_num; /* slo obrazovky, kterou bude program pouvat */
Screen *screen; /* Struktura popisujc obrazovku */
unsigned int screen_width, screen_height; /* Rozmry obrazovky v pixelech */
unsigned int screen_width_mm, screen_height_mm; /* Rozmry obrazovky v mm */
Window root; /* Root okno */
unsigned long black, white; /* Hodnoty BlackPixel a WhitePixel */


Window topwin, win1, win2; /* Okna */
Pixmap topwin_icon; /* Pixmapa pro ikonu top-level okna */
int topwin_x, topwin_y; /* Pozice okna */
unsigned int topwin_w, topwin_h, topwin_b; /* Rozmry okna a okraje */

GC def_gc; /* Standardn grafick kontext */
GC my_gc; /* Grafick kontext, vytvoen se standardnmi hodnotami */
XFontStruct *def_font; /* Informace o standardnm fontu */

int do_debug = 1; /* Vypisovat ladic hlky */

/*
 * Funkce pro vpis ladicch hlek
 */
static void debug(const char *msg, ...)
{
    va_list args;
    if(!do_debug)
	return;
    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
}

/*
 * Pipojit k serveru, zjistit o nm informace
 */
static void open_display(void)
{
    display_name = XDisplayName(NULL);
    if(!(display = XOpenDisplay(display_name))) {
	fprintf(stderr, "Cannot connect to X server \"%s\"\n", display_name);
	exit(1);
    }
    debug("Connected to X server \"%s\"\n", display_name);
    
    screen_num = DefaultScreen(display);
    screen = DefaultScreenOfDisplay(display);
    debug("Default screen number is %d\n", screen_num);
    
    screen_width = DisplayWidth(display, screen_num);
    screen_height = DisplayHeight(display, screen_num);
    screen_width_mm = DisplayWidthMM(display, screen_num);
    screen_height_mm = DisplayHeightMM(display, screen_num);
    debug("Screen size is %ux%u pixels, %ux%u mm\n", screen_width,
	  screen_height, screen_width_mm, screen_height_mm);
    
    root = RootWindow(display, screen_num);
    black = BlackPixel(display, screen_num);
    white = WhitePixel(display, screen_num);
    debug("Root window ID is %#lx, black is %#lx, white is %#lx\n", root,
	  black, white);

    def_gc = DefaultGC(display, screen_num);
    if(!(def_font = XQueryFont(display, XGContextFromGC(def_gc)))) {
	fprintf(stderr, "XQueryFont() failed\n");
	exit(1);
    }
}

/*
 * Vytvoit a namapovat top-level okno, vytvoit dal zdroje (nap. GC)
 */
static void create_resources(int argc, char *argv[])
{
    XWMHints *wm_hints; /* Doporuen pro window manager */
    XSizeHints *size_hints; /* Doporuen velikost okna */
    XClassHint *class_hint; /* Jmno a tda pro resource manager */
    XTextProperty window_name, icon_name; /* Jmna pro window manager */
    XGCValues gc_values; /* Hodnoty pro nastaven GC */
    char *str_window_name = (char *)"Hello Window"; /* Titulek okna */
    char *str_icon_name = (char *)"Hello Icon"; /* Titulek ikony */
    Cursor cursor; /* Kurzor */
    XSetWindowAttributes set_attr;
    XColor bg, fg; /* Barvy pro kurzor */
    Pixmap source, mask; /* Bitmapy pro kurzor */
    int cw, ch; /* Velikost kurzoru */
    int qbc; /* Nvratov hodnota XQueryBestCursor() */
    Font font; /* Font pro vytvoen kurzoru */
   
    /* Vytvoit top-level okno */
    topwin_x = topwin_y = 0; /* Okno umst window manager */
    topwin_w = screen_width / 2;
    topwin_h = screen_height / 2;
    topwin_b = 0;
    topwin = XCreateSimpleWindow(display, root, topwin_x, topwin_y, topwin_w,
				 topwin_h, topwin_b, black, white);
    debug("Created top-level window ID %#lx\n", topwin);

    /* Pipravit barvy pro kurzor */
    fg.red = 0x8000;
    fg.green = fg.blue = 0;
    bg.green = 0xffff;
    bg.red = bg.blue = 0;
    
    /* Nastavit kurzor top-level okna */
    cursor = XCreateFontCursor(display, XC_cross);
    XRecolorCursor(display, cursor, &fg, &bg);
    XDefineCursor(display, topwin, cursor);

    /* Vytvoit ikonu pro top-level okno */
    topwin_icon = XCreateBitmapFromData(display, topwin, Xlib_icon_bits,
					Xlib_icon_width, Xlib_icon_height);
    debug("Created icon %#lx for top-level window\n", topwin_icon);

    /* Nastavit properties pro window manager */
    if(!(size_hints = XAllocSizeHints()) ||
       !(wm_hints = XAllocWMHints()) ||
       !(class_hint = XAllocClassHint())) {
	fprintf(stderr, "Cannot allocate memory\n");
	exit(1);
    }
    
    /*  Poloha a velikost se bere ze skutench hodnot okna v okaiku
     * namapovn. PPosition | PSize k, e hodnoty nastavil program (a
     * window manager je me mnit podle poteby). USSize | USPosition by
     * znamenalo, e hodnoty zadal uivatel (nap. na pkazov dce) a
     * window manager by je neml mnit. */
    size_hints->flags = PPosition | PSize | PMinSize;
    /* Window manager by neml okno nikdy zmenit na mn ne
     * min_width x min_height pixel. */
    size_hints->min_width = 200;
    size_hints->min_height = 100;
    
    wm_hints->initial_state = NormalState;/* Na zatku m bt zobrazeno
                                             normln okno (ne ikona) */
    wm_hints->input = True; /* Aplikace potebuje vstup z klvesnice */
    wm_hints->icon_pixmap = topwin_icon;
    wm_hints->flags = StateHint | InputHint | IconPixmapHint;
    
    /* Uloen jmna okna a ikony */
    if(!XStringListToTextProperty(&str_window_name, 1, &window_name) ||
       !XStringListToTextProperty(&str_icon_name, 1, &icon_name)) {
        fprintf(stderr, "XStringListToTextProperty() for failed\n");
        exit(1);
    }

    /* Jmno programu a tdy pro hledn v resource databzi */
    class_hint->res_name = progname;
    class_hint->res_class = (char *) "HelloWorld";

    /* Nastavit vechny properties */
    XSetWMProperties(display, topwin, &window_name, &icon_name, argv, argc,
                     size_hints, wm_hints, class_hint);

    /* Vybrat udlosti pro top-level okno */
    XSelectInput(display, topwin, KeyPressMask | ButtonPressMask);
    
    /* Vytvoit dal okna */
    qbc = XQueryBestCursor(display, topwin, cursor_source_width, 
	      		   cursor_source_height, &cw, &ch);
    debug("XQueryBestCursor(%d, %d) = %d -> %dx%d\n", cursor_source_width,
	  cursor_source_height, qbc, cw, ch);
    source = XCreateBitmapFromData(display, topwin, cursor_source_bits,
			   	   cursor_source_width, cursor_source_height);
    mask = XCreateBitmapFromData(display, topwin, cursor_mask_bits,
				 cursor_mask_width, cursor_mask_height);
    set_attr.cursor = XCreatePixmapCursor(display, source, mask, &fg, &bg,
					  4, 4);
    win1 = XCreateWindow(display, topwin, 20, 10, 150, 75, 0, CopyFromParent,
			 InputOnly, CopyFromParent, CWCursor, &set_attr);

    font = XLoadFont(display, "fixed");
    set_attr.cursor = XCreateGlyphCursor(display, font, font, '+', '*',
					 &fg, &bg);
    win2 = XCreateWindow(display, topwin, 200, 10, 150, 75, 0, CopyFromParent,
			 InputOnly, CopyFromParent, CWCursor, &set_attr);
    
    /* Vytvoit GC */
    gc_values.background = white;
    gc_values.foreground = black;
    my_gc = XCreateGC(display, topwin, GCBackground | GCForeground, &gc_values);
    XCopyGC(display, def_gc, GCFont, my_gc); /* default font */
    
    /* Namapovat okna */
    XMapWindow(display, win1);
    XMapWindow(display, win2);
    XMapWindow(display, topwin);
}

/*
 * Hlavn cyklus zpracovn udlost
 */
static void event_loop(void)
{
    XEvent event;
    char buf[10];
    KeySym keysym;
    
    while(1) {
	XNextEvent(display, &event);
	switch(event.type) {
	    case ButtonPress:
		XDrawPoint(display, topwin, def_gc, event.xbutton.x,
			   event.xbutton.y);
		break;
	    case KeyPress:
		XLookupString(&event.xkey, buf, 10, &keysym, NULL);
		debug("Key \"%s\" pressed\n", XKeysymToString(keysym));
		return;
		break;
	    case MappingNotify:
		/* Nast zmnn mapovn klves */
		if(event.xmapping.request == MappingKeyboard)
		    XRefreshKeyboardMapping(&event.xmapping);
		break;
	    default:
		/* Zbyl nezajmav udlosti */
		break;
	}
    }
}

/*
 * Odpojit od X serveru
 */
static void close_display(void)
{
    /* Po odpojen klienta X server automaticky zru vechny jm vytvoen
     * zdroje. XCloseDisplay() volme hlavn proto, abychom dostali ppadn
     * chyby, kter jsem zatm nedostali. */
    XCloseDisplay(display);
}

/*
 * Vstupn bod programu
 */
int main(int argc, char *argv[])
{
    progname = argv[0];

    open_display();
    create_resources(argc, argv);
    event_loop();
    close_display();
   
    return 0;
}
