/* Ukzka styl ar */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "Xlib_icon.h"

/*
 * Globln pouvan promnn
 */
char *progname; /* Jmno programu z argv[0] */

char *display_name; /* Jmno pouvanho displeje */
Display *display; /* Struktura popisujc displej */
int screen_num; /* slo obrazovky, kterou bude program pouvat */
Screen *screen; /* Struktura popisujc obrazovku */
unsigned int screen_width, screen_height; /* Rozmry obrazovky v pixelech */
unsigned int screen_width_mm, screen_height_mm; /* Rozmry obrazovky v mm */
Window root; /* Root okno */
unsigned long black, white; /* Hodnoty BlackPixel a WhitePixel */
Colormap cmap; /* Standardn paleta */

Window topwin; /* Top-level okno */
Pixmap topwin_icon; /* Pixmapa pro ikonu top-level okna */
int topwin_x, topwin_y; /* Pozice okna */
unsigned int topwin_w, topwin_h, topwin_b; /* Rozmry okna a okraje */

GC def_gc; /* Standardn grafick kontext */
GC my_gc; /* Grafick kontext, vytvoen se standardnmi hodnotami */
XFontStruct *def_font; /* Informace o standardnm fontu */

int do_debug = 1; /* Vypisovat ladic hlky */

/*
 * Funkce pro vpis ladicch hlek
 */
static void debug(const char *msg, ...)
{
    va_list args;
    if(!do_debug)
	return;
    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
}

/*
 * Pipojit k serveru, zjistit o nm informace
 */
static void open_display(void)
{
    display_name = XDisplayName(NULL);
    if(!(display = XOpenDisplay(display_name))) {
	fprintf(stderr, "Cannot connect to X server \"%s\"\n", display_name);
	exit(1);
    }
    debug("Connected to X server \"%s\"\n", display_name);
    
    screen_num = DefaultScreen(display);
    screen = DefaultScreenOfDisplay(display);
    debug("Default screen number is %d\n", screen_num);
    
    screen_width = DisplayWidth(display, screen_num);
    screen_height = DisplayHeight(display, screen_num);
    screen_width_mm = DisplayWidthMM(display, screen_num);
    screen_height_mm = DisplayHeightMM(display, screen_num);
    debug("Screen size is %ux%u pixels, %ux%u mm\n", screen_width,
	  screen_height, screen_width_mm, screen_height_mm);
    
    root = RootWindow(display, screen_num);
    black = BlackPixel(display, screen_num);
    white = WhitePixel(display, screen_num);
    cmap = DefaultColormap(display, screen_num);
    debug("Root window ID is %#lx, black is %#lx, white is %#lx\n", root,
	  black, white);

    def_gc = DefaultGC(display, screen_num);
    /* Toto voln XQueryFont() do def_font->fid neulo ID fontu, ale
     * grafickho kontextu! Nevm, jak zskat ID nebo jmno standardnho
     * fontu. */
    if(!(def_font = XQueryFont(display, XGContextFromGC(def_gc)))) {
	fprintf(stderr, "XQueryFont() failed\n");
	exit(1);
    }
}

/*
 * Vytvoit a namapovat top-level okno, vytvoit dal zdroje (nap. GC)
 */
static void create_resources(int argc, char *argv[])
{
    XWMHints *wm_hints; /* Doporuen pro window manager */
    XSizeHints *size_hints; /* Doporuen velikost okna */
    XClassHint *class_hint; /* Jmno a tda pro resource manager */
    XTextProperty window_name, icon_name; /* Jmna pro window manager */
    XGCValues gc_values; /* Hodnoty pro nastaven GC */
    char *str_window_name = (char *)"Hello Window"; /* Titulek okna */
    char *str_icon_name = (char *)"Hello Icon"; /* Titulek ikony */
    XColor exact_bg, screen_bg; /* Barva pozad pro GC */
   
    /* Vytvoit top-level okno */
    topwin_x = topwin_y = 0; /* Okno umst window manager */
    topwin_w = screen_width / 2;
    topwin_h = screen_height / 2;
    topwin_b = 0;
    topwin = XCreateSimpleWindow(display, root, topwin_x, topwin_y, topwin_w,
				 topwin_h, topwin_b, black, white);
    debug("Created top-level window ID %#lx\n", topwin);

    /* Vytvoit ikonu pro top-level okno */
    topwin_icon = XCreateBitmapFromData(display, topwin, Xlib_icon_bits,
					Xlib_icon_width, Xlib_icon_height);
    debug("Created icon %#lx for top-level window\n", topwin_icon);

    /* Nastavit properties pro window manager */
    if(!(size_hints = XAllocSizeHints()) ||
       !(wm_hints = XAllocWMHints()) ||
       !(class_hint = XAllocClassHint())) {
	fprintf(stderr, "Cannot allocate memory\n");
	exit(1);
    }
    
    /*  Poloha a velikost se bere ze skutench hodnot okna v okaiku
     * namapovn. PPosition | PSize k, e hodnoty nastavil program (a
     * window manager je me mnit podle poteby). USSize | USPosition by
     * znamenalo, e hodnoty zadal uivatel (nap. na pkazov dce) a
     * window manager by je neml mnit. */
    size_hints->flags = PPosition | PSize | PMinSize;
    /* Window manager by neml okno nikdy zmenit na mn ne
     * min_width x min_height pixel. */
    size_hints->min_width = 200;
    size_hints->min_height = 100;
    
    wm_hints->initial_state = NormalState;/* Na zatku m bt zobrazeno
                                             normln okno (ne ikona) */
    wm_hints->input = True; /* Aplikace potebuje vstup z klvesnice */
    wm_hints->icon_pixmap = topwin_icon;
    wm_hints->flags = StateHint | InputHint | IconPixmapHint;
    
    /* Uloen jmna okna a ikony */
    if(!XStringListToTextProperty(&str_window_name, 1, &window_name) ||
       !XStringListToTextProperty(&str_icon_name, 1, &icon_name)) {
        fprintf(stderr, "XStringListToTextProperty() for failed\n");
        exit(1);
    }

    /* Jmno programu a tdy pro hledn v resource databzi */
    class_hint->res_name = progname;
    class_hint->res_class = (char *) "HelloWorld";

    /* Nastavit vechny properties */
    XSetWMProperties(display, topwin, &window_name, &icon_name, argv, argc,
                     size_hints, wm_hints, class_hint);

    /* Vybrat udlosti pro top-level okno */
    XSelectInput(display, topwin, ExposureMask | KeyPressMask |
		 ButtonPressMask | StructureNotifyMask);
    
    /* Vytvoit GC */
    if(!XAllocNamedColor(display, cmap, "RGBi:0.5/0.5/0.5", &exact_bg,
			 &screen_bg)) {
	fprintf(stderr, "Cannot allocate color\n");
	exit(1);
    }
    gc_values.background = screen_bg.pixel;
    gc_values.foreground = black;
    my_gc = XCreateGC(display, topwin, GCBackground | GCForeground, &gc_values);
    XCopyGC(display, def_gc, GCFont, my_gc); /* default font */
    
    /* Namapovat top-level okno */
    XMapWindow(display, topwin);
}

/*
 * Kreslen
 */
static void expose(XExposeEvent *event)
{
    static char dash_list1[] = {15, 20, 15, 5}; /* Sud poet poloek */
    static char dash_list2[] = {5, 10, 15}; /* Lich poet poloek */
   
    if(event->count > 0)
	return; /* Kreslit jen po posledn udlosti v ad */
    
    XSetLineAttributes(display, my_gc, 10, LineSolid, CapButt, JoinRound);
    XDrawLine(display, topwin, my_gc, 10, 20, 310, 20);
    XDrawString(display, topwin, def_gc, 320, 20, "LineSolid", 9);
    
    XSetLineAttributes(display, my_gc, 10, LineOnOffDash, CapButt, JoinRound);
    XSetDashes(display, my_gc, 0, dash_list1, 4);
    XDrawLine(display, topwin, my_gc, 10, 50, 310, 50);
    XDrawString(display, topwin, def_gc, 320, 50, "LineOnOffDash", 13);
    
    XSetDashes(display, my_gc, 10, dash_list1, 4);
    XSetLineAttributes(display, my_gc, 10, LineDoubleDash, CapButt, JoinRound);
    XDrawLine(display, topwin, my_gc, 10, 80, 310, 80);
    XDrawString(display, topwin, def_gc, 320, 80, "LineDoubleDash", 14);
    
    XSetLineAttributes(display, my_gc, 10, LineOnOffDash, CapButt, JoinRound);
    XSetDashes(display, my_gc, 0, dash_list2, 3);
    XDrawLine(display, topwin, my_gc, 10, 110, 310, 110);
    XDrawString(display, topwin, def_gc, 320, 110, "LineOnOffDash", 13);
}

/*
 * Hlavn cyklus zpracovn udlost
 */
static void event_loop(void)
{
    XEvent event;
    char buf[10];
    KeySym keysym;
    
    while(1) {
	XNextEvent(display, &event);
	switch(event.type) {
	    case Expose:
		expose(&event.xexpose);
		break;
	    case ButtonPress:
		debug("Button %u pressed\n", event.xbutton.button);
		return;
		break;
	    case KeyPress:
		XLookupString(&event.xkey, buf, 10, &keysym, NULL);
		debug("Key \"%s\" pressed\n", XKeysymToString(keysym));
		return;
		break;
	    case ConfigureNotify:
		if((int)topwin_w != event.xconfigure.width ||
		   (int)topwin_h != event.xconfigure.height) {
		    topwin_w = event.xconfigure.width;
		    topwin_h = event.xconfigure.height;
		    debug("Window resized to %ux%u pixels\n",
			  topwin_w, topwin_h);
		    XClearWindow(display, topwin);
		}
		break;
	    case MappingNotify:
		/* Nast zmnn mapovn klves */
		if(event.xmapping.request == MappingKeyboard)
		    XRefreshKeyboardMapping(&event.xmapping);
		break;
	    default:
		/* Zbyl nezajmav udlosti */
		break;
	}
    }
}

/*
 * Odpojit od X serveru
 */
static void close_display(void)
{
    /* Po odpojen klienta X server automaticky zru vechny jm vytvoen
     * zdroje. XCloseDisplay() volme hlavn proto, abychom dostali ppadn
     * chyby, kter jsem zatm nedostali. */
    XCloseDisplay(display);
}

/*
 * Vstupn bod programu
 */
int main(int argc, char *argv[])
{
    progname = argv[0];

    open_display();
    create_resources(argc, argv);
    event_loop();
    close_display();
   
    return 0;
}
